"""Test of data set functionalities"""
import numpy as np
import pytest

import pyjibe.head

from helpers import make_directory_with_data


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = pyjibe.head.PyJibe()
    main_window.close()


def test_clear_and_verify_data(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data())
    war = main_window.subwindows[0].widget()
    # clear data
    tpp = war.tab_preprocess
    war.cb_autosave.setChecked(0)
    # perform simple filter
    tpp.set_preprocessing(["compute_tip_position"])
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # disable weighting
    war.tab_fit.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    apret = war.data_set[0]
    war.tab_fit.fit_approach_retract(apret)
    # Now check something
    ftab = war.tab_fit.table_parameters_fitted
    # E
    assert np.allclose(apret.fit_properties["params_fitted"]["E"].value,
                       14741.958242422093,
                       atol=1e-4,
                       rtol=0)
    assert float(ftab.item(0, 0).text()) == 14742
    # contact_point
    assert float(ftab.item(1, 0).text()) == 18029
    # baseline_offset
    assert float(ftab.item(2, 0).text()) == -480.67


def test_fit_all(qtbot):
    """Perform a simple fit with the standard parameters"""
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    war.cb_autosave.setChecked(0)
    war.on_fit_all()
    a1 = war.data_set[0]
    a2 = war.data_set[1]
    assert a1.fit_properties == a2.fit_properties


@pytest.mark.parametrize("method,contact_point", [
    ["gradient_zero_crossing", 1895],
    ["fit_constant_line", 1919],
    ["deviation_from_baseline", 1908],
    ])
def test_preprocessing_poc_estimation(method, contact_point):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    # clear data
    war.cb_autosave.setChecked(0)
    war.tabs.setCurrentIndex(0)
    # perform simple filter
    war.tab_preprocess.set_preprocessing(
        preprocessing=["compute_tip_position", "correct_tip_offset"],
        options={"correct_tip_offset": {"method": method}}
    )
    war.tabs.setCurrentIndex(1)
    fd = war.data_set[0]
    assert np.argmin(np.abs(fd["tip position"])) == contact_point



from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)

from ..head import custom_widgets


class MPLPreproc:
    def __init__(self):
        """Matplotlib plot for preprocessing data"""
        # Add matplotlib figure
        self.figure = Figure(facecolor="none", tight_layout=True,
                             frameon=True)
        self.axis = self.figure.add_subplot(111)
        self.axis.set_facecolor('#FFFFFF')
        self.canvas = FigureCanvas(self.figure)
        self.canvas.draw()

    def add_toolbar(self, widget):
        """Add toolbar to PyQT widget"""
        self.toolbar = custom_widgets.NavigationToolbarPreproc(
            self.canvas,
            widget,
            coordinates=True
        )

    def clear(self):
        self.axis.clear()
        self.axis.grid()



import io
import pathlib
import shutil
import sys
import tempfile
from unittest import mock

import pytest

import pyjibe
import pyjibe.head


data_path = pathlib.Path(__file__).parent / "data"


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python>=3.8")
def test_init_print_version(qtbot):
    """I just wanted to show off how good I am with mock and pytest"""
    mock_stdout = io.StringIO()
    mock_exit = mock.Mock()

    with mock.patch("sys.argv", ["", "--version"]):
        with mock.patch("sys.exit", mock_exit):
            with mock.patch('sys.stdout', mock_stdout):
                mw = pyjibe.head.PyJibe()
                mw.close()

    assert mock_exit.call_args.args[0] == 0
    assert mock_stdout.getvalue().strip() == pyjibe.__version__


def test_rem_subwindow(qtbot):
    def mock_get_open_filenames(*args, **kwargs):
        tdir = tempfile.mkdtemp(prefix="pyjibe_test_open_bulk_")
        t2 = pathlib.Path(tdir) / "spot1.jpk-force"
        shutil.copy2(data_path / "spot3-0192.jpk-force", t2)
        return [t2], None

    mw = pyjibe.head.PyJibe()

    with mock.patch("PyQt5.QtWidgets.QFileDialog.getOpenFileNames",
                    mock_get_open_filenames):
        mw.on_open_single()

    sub = mw.subwindows[0]
    title = sub.windowTitle()
    action_texts = [action.text() for action in mw.menuExport.actions()]
    assert title in action_texts
    sub.close()
    action_texts_2 = [action.text() for action in mw.menuExport.actions()]
    assert title not in action_texts_2
    assert len(mw.subwindows) == 0
    mw.close()


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python>=3.8")
def test_on_about(qtbot):
    with mock.patch("PyQt5.QtWidgets.QMessageBox.about") as mock_about:
        mw = pyjibe.head.PyJibe()
        mw.on_about()
        mw.close()

        assert mock_about.call_args.args[1] == f"PyJibe {pyjibe.__version__}"
        assert "PyJibe" in mock_about.call_args.args[2]


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python>=3.8")
def test_on_documentation(qtbot):
    with mock.patch("webbrowser.open") as mock_wbopen:
        mw = pyjibe.head.PyJibe()
        mw.on_documentation()
        mw.close()

        assert mock_wbopen.call_args.args[0] == "https://pyjibe.readthedocs.io"


def test_on_open_bulk(qtbot):
    def mock_selected_files(self):
        tdir = tempfile.mkdtemp(prefix="pyjibe_test_open_bulk_")
        t2 = pathlib.Path(tdir) / "spot1.jpk-force"
        t3 = pathlib.Path(tdir) / "spot2.jpk-force"
        t4 = pathlib.Path(tdir) / "spot3.jpk-force"
        shutil.copy2(data_path / "spot3-0192.jpk-force", t2)
        shutil.copy2(data_path / "spot3-0192.jpk-force", t3)
        shutil.copy2(data_path / "spot3-0192.jpk-force", t4)
        return [t2, t3, t4]

    mw = pyjibe.head.PyJibe()

    with mock.patch("pyjibe.head.custom_widgets."
                    "DirectoryDialogMultiSelect.selectedFiles",
                    mock_selected_files):
        with mock.patch("pyjibe.head.custom_widgets."
                        "DirectoryDialogMultiSelect.exec_"):
            mw.on_open_bulk()

    assert len(mw.subwindows) == 1
    war = mw.subwindows[0].widget()
    assert war.list_curves.topLevelItemCount() == 3
    mw.close()


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python>=3.8")
def test_on_open_bulk_no_data(qtbot):
    def mock_selected_files(self):
        return [tempfile.mkdtemp(prefix="pyjibe_no_data_")]

    mw = pyjibe.head.PyJibe()

    with mock.patch("pyjibe.head.custom_widgets."
                    "DirectoryDialogMultiSelect.selectedFiles",
                    mock_selected_files):
        with mock.patch("pyjibe.head.custom_widgets."
                        "DirectoryDialogMultiSelect.exec_"):
            with mock.patch("PyQt5.QtWidgets.QMessageBox.warning") as warn:
                mw.on_open_bulk()
                assert warn.call_args.args[1] == "No AFM data found!"

    mw.close()


def test_on_open_multiple(qtbot):
    def mock_selected_files(self):
        tdir = tempfile.mkdtemp(prefix="pyjibe_test_open_bulk_")
        t2 = pathlib.Path(tdir) / "1" / "spot1.jpk-force"
        t3 = pathlib.Path(tdir) / "2" / "spot2.jpk-force"
        t4 = pathlib.Path(tdir) / "3" / "spot3.jpk-force"
        t2.parent.mkdir()
        t3.parent.mkdir()
        t4.parent.mkdir()
        shutil.copy2(data_path / "spot3-0192.jpk-force", t2)
        shutil.copy2(data_path / "spot3-0192.jpk-force", t3)
        shutil.copy2(data_path / "spot3-0192.jpk-force", t4)
        return [t2, t3, t4]

    mw = pyjibe.head.PyJibe()

    with mock.patch("pyjibe.head.custom_widgets."
                    "DirectoryDialogMultiSelect.selectedFiles",
                    mock_selected_files):
        with mock.patch("pyjibe.head.custom_widgets."
                        "DirectoryDialogMultiSelect.exec_"):
            mw.on_open_multiple()

    assert len(mw.subwindows) == 3
    war = mw.subwindows[0].widget()
    assert war.list_curves.topLevelItemCount() == 1
    mw.close()


def test_on_open_single(qtbot):
    def mock_get_open_filenames(*args, **kwargs):
        tdir = tempfile.mkdtemp(prefix="pyjibe_test_open_bulk_")
        t2 = pathlib.Path(tdir) / "spot1.jpk-force"
        shutil.copy2(data_path / "spot3-0192.jpk-force", t2)
        return [t2], None

    mw = pyjibe.head.PyJibe()

    with mock.patch("PyQt5.QtWidgets.QFileDialog.getOpenFileNames",
                    mock_get_open_filenames):
        mw.on_open_single()

    assert len(mw.subwindows) == 1
    war = mw.subwindows[0].widget()
    assert war.list_curves.topLevelItemCount() == 1
    mw.close()


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python>=3.8")
def test_on_software(qtbot):
    mw = pyjibe.head.PyJibe()

    with mock.patch("PyQt5.QtWidgets.QMessageBox.information") as info:
        mw.on_software()
        assert info.call_args.args[1] == "Software"
        assert info.call_args.args[2].count(f"PyJibe {pyjibe.__version__}")
        assert info.call_args.args[2].count("lmfit")

    mw.close()



"""Test built-in rater"""
import pathlib
import shutil
import tempfile
from unittest import mock

import h5py
from PyQt5 import QtCore

import pyjibe.head


data_path = pathlib.Path(__file__).parent / "data"


def test_rater_basic(qtbot):
    mw = pyjibe.head.PyJibe()
    td = pathlib.Path(tempfile.mkdtemp(prefix="rater_")) / "map.jpk-force-map"
    shutil.copy2(data_path / "map2x2_extracted.jpk-force-map", td)
    mw.load_data([td])

    war = mw.subwindows[0].widget()

    h5out = pathlib.Path(tempfile.mkdtemp(prefix="rate_")) / "rate.h5"
    with mock.patch("PyQt5.QtWidgets.QFileDialog.getSaveFileName",
                    lambda *args, **kwargs: (str(h5out), None)):
        war.on_user_rate()

    rater = war.curve_rater  # it's not a modal dialog

    # fill in a few values
    assert rater.sp_rating.value() == -1
    rater.sp_rating.setValue(8)
    qtbot.mouseClick(rater.btn_next, QtCore.Qt.LeftButton)
    assert rater.curve_index.value() == 2
    assert rater.sp_rating.value() == -1
    rater.text_comment.setPlainText("peter")
    qtbot.mouseClick(rater.btn_prev, QtCore.Qt.LeftButton)
    assert rater.curve_index.value() == 1
    assert rater.sp_rating.value() == 8
    assert rater.text_comment.toPlainText() == ""
    qtbot.mouseClick(rater.btn_next, QtCore.Qt.LeftButton)
    qtbot.mouseClick(rater.btn_next, QtCore.Qt.LeftButton)
    assert rater.curve_index.value() == 3
    assert rater.sp_rating.value() == -1
    mw.close()

    with h5py.File(h5out, "r") as h5:
        assert len(h5["analysis"]) == 2
        keys = list(h5["analysis"].keys())
        assert h5["analysis"][keys[0]].attrs["user rate"] == 8
        assert h5["analysis"][keys[1]].attrs["user comment"] == "peter"



"""Test of data set functionalities"""
import pathlib
import shutil
import tempfile

import nanite.model as nmodel
import numpy as np
import pytest
from PyQt5 import QtCore, QtWidgets

import pyjibe.head

from helpers import make_directory_with_data


data_path = pathlib.Path(__file__).parent / "data"


class MockModel:
    def __init__(self, model_key, **kwargs):
        # rebase on hertz model
        md = nmodel.models_available["hertz_para"]
        for key in dir(md):
            setattr(self, key, getattr(md, key))
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])
        self.model_key = model_key

    def __enter__(self):
        nmodel.register_model(self, self.__repr__())

    def __exit__(self, a, b, c):
        nmodel.models_available.pop(self.model_key)


def test_ancillary_update_init(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {
            # take initial fit parameter of E
            "E": x.get_initial_fit_parameters(
                model_ancillaries=False)["E"].value},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        war.on_tab_changed()
        assert len(war.data_set[0].preprocessing) == 1
        assert len(war.tab_preprocess.current_preprocessing()[0]) == 1
        # The ancillary parameter gets its value from the default parameters
        assert atab.item(0, 1).text() == "3000"
        assert itab.item(0, 1).text() == "3000"
        # Now we change the initial parameter "E" and move on to the next
        # curve. Ancillary parameter "F" should also change.
        itab.item(0, 1).setText("2000")
        assert atab.item(0, 1).text() == "2000"
        it = war.list_curves.topLevelItem(1)
        war.list_curves.setCurrentItem(it)
        assert itab.item(0, 1).text() == "2000"
        assert atab.item(0, 1).text() == "2000"


def test_ancillary_update_nan(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {"E": np.nan},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        assert atab.item(0, 1).text() == "nan"
        assert itab.item(0, 1).text() == "3000"


def test_ancillary_update_preproc_change(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {
            # i.e. model works only if there are multiple preproc steps
            "E": np.nan if len(x.preprocessing) == 1 else 2345},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        war.on_tab_changed()
        assert len(war.data_set[0].preprocessing) == 1
        assert len(war.tab_preprocess.current_preprocessing()[0]) == 1
        assert atab.item(0, 1).text() == "nan"
        assert itab.item(0, 1).text() == "3000"
        # up until here this is the same as `test_update_ancillary_nan`
        # now change preprocessing
        war.tabs.setCurrentIndex(0)  # actually switch tabs like a user
        # manually check the "correct_tip_offset" widget
        for pwid, k in war.tab_preprocess._map_widgets_to_preproc_ids.items():
            if k == "correct_tip_offset":
                pwid.setChecked(True)
        war.tabs.setCurrentIndex(1)  # triggers recomputation of anc
        assert len(war.data_set[0].preprocessing) == 2
        assert len(war.tab_preprocess.current_preprocessing()[0]) == 2
        assert atab.item(0, 1).text() == "2345"
        assert itab.item(0, 1).text() == "2345"


@pytest.mark.filterwarnings('ignore::UserWarning')
def test_apply_and_fit_all_with_bad_data(qtbot, monkeypatch):
    # setup data directory with two good and one invalid file
    td = pathlib.Path(tempfile.mkdtemp(prefix="pyjibe_test_apply_fit_all_"))
    shutil.copy2(data_path / "spot3-0192.jpk-force", td / "data1.jpk-force")
    shutil.copy2(data_path / "invalid_dataset.jpk-force",
                 td / "data2.jpk-force")
    shutil.copy2(data_path / "spot3-0192.jpk-force", td / "data3.jpk-force")
    files = sorted(td.glob("*.jpk-force"))
    # sanity checks
    assert len(files) == 3
    assert files[1].name == "data2.jpk-force"

    # monkeypatch message dialog
    message_list = []
    monkeypatch.setattr(
        QtWidgets.QMessageBox, "warning",
        lambda parent, title, message: message_list.append(message))

    # initialize
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=files)
    war = main_window.subwindows[0].widget()
    war.tab_preprocess.set_preprocessing(
        ["compute_tip_position", "correct_force_offset", "correct_tip_offset"])

    # Hit "apply model and fit all"
    qtbot.mouseClick(war.btn_fitall, QtCore.Qt.LeftButton, delay=200)

    # make sure that we got that message
    assert message_list
    assert "data2.jpk-force" in message_list[0]

    # make sure the curves got rated
    good1 = war.list_curves.topLevelItem(0)
    assert float(good1.data(2, 0)) > 0  # column 2 shows the rating
    bad = war.list_curves.topLevelItem(1)
    assert float(bad.data(2, 0)) == -1  # column 2 shows the rating
    good2 = war.list_curves.topLevelItem(2)
    assert float(good2.data(2, 0)) > 0  # column 2 shows the rating


def test_change_model_keep_parms(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    # clear data
    war.cb_autosave.setChecked(0)
    # perform simple filter
    war.tab_preprocess.set_preprocessing(["compute_tip_position"])
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # set value for contact point
    itab.item(3, 1).setText(str(12345))
    # change the model to pyramidal
    pyr_name = nmodel.model_hertz_three_sided_pyramid.model_name
    pyr_idx = war.tab_fit.cb_model.findText(pyr_name)
    war.tab_fit.cb_model.setCurrentIndex(pyr_idx)
    # check that contact point is still the same
    assert float(itab.item(3, 1).text()) == 12345


def test_remember_initial_params(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    # clear data
    war.cb_autosave.setChecked(0)
    # perform simple filter
    war.tab_preprocess.set_preprocessing(["compute_tip_position"])
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # disable weighting
    war.tab_fit.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    # change standard tip radius from 10 to 5
    assert float(itab.item(1, 1).text()) == 10
    itab.item(1, 1).setText(str(5))
    cl1 = war.list_curves.currentItem()
    cl2 = war.list_curves.itemBelow(cl1)
    war.list_curves.setCurrentItem(cl2)
    assert float(itab.item(1, 1).text()) == 5


def test_set_indentation_depth_manually_infdoublespinbox(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # set value for contact point
    itab.item(3, 1).setText(str(12345))
    # change the model to pyramidal
    pyr_name = nmodel.model_hertz_three_sided_pyramid.model_name
    pyr_idx = war.tab_fit.cb_model.findText(pyr_name)
    war.tab_fit.cb_model.setCurrentIndex(pyr_idx)
    # set left fitting range
    for text_entered, resulting_value in [
        ["-1.40", -1.4],
        ["1...2", 1.2],
        ["1.0e-4", 1e-4],
        ["inf", np.inf],
        ["1.10201", 1.10201],
        ["1.001", 1.001],
        ["-1.04", -1.04]
            ]:
        war.tab_fit.sp_range_1.clear()
        qtbot.keyClicks(war.tab_fit.sp_range_1, text_entered)
        assert war.tab_fit.sp_range_1.value() == resulting_value



import pathlib

from PyQt5 import QtCore, QtWidgets


class DirectoryDialogMultiSelect(QtWidgets.QFileDialog):
    def __init__(self, *args):
        """A dialog that lets the user select multiple directories"""
        QtWidgets.QFileDialog.__init__(self, *args)
        self.setOption(self.DontUseNativeDialog, True)
        self.setFileMode(self.DirectoryOnly)

        self.tree = self.findChild(QtWidgets.QTreeView)
        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)

        self.list = self.findChild(QtWidgets.QListView)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)

        for view in self.findChildren((QtWidgets.QListView,
                                       QtWidgets.QTreeView)):
            if isinstance(view.model(), QtWidgets.QFileSystemModel):
                view.setSelectionMode(
                    QtWidgets.QAbstractItemView.MultiSelection)

        # Add common directories in the sidebar
        sidebar_urls = self.sidebarUrls()
        for path in ["/Volumes", "/media", "/mnt", "/"]:
            if pathlib.Path(path).exists():
                sidebar_urls.append(QtCore.QUrl('file://' + path))
        self.setSidebarUrls(sorted(set(sidebar_urls)))

    def getDirectory(self):
        """Return a string of the root directory"""
        files = self.selectedFiles()
        if files:
            return str(pathlib.Path(files[0]).parent)
        else:
            return None



import numpy as np

from PyQt5 import QtWidgets, QtGui


class InfDoubleSpinBox(QtWidgets.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        super(InfDoubleSpinBox, self).__init__(*args, **kwargs)
        self.setMinimum(-np.inf)
        self.setMaximum(np.inf)
        self.validator = FloatValidator()
        self.setDecimals(1000)
        self.setSingleStep(0.1)

    def validate(self, text, position):
        return self.validator.validate(text, position)

    def fixup(self, text):
        return self.validator.fixup(text)

    def valueFromText(self, text):
        return convert_string_to_float(text)

    def textFromValue(self, value):
        return convert_float_to_string(value)


class FloatValidator(QtGui.QValidator):
    def __init__(self, *args, **kwargs):
        super(FloatValidator, self).__init__(*args, **kwargs)
        self.previous = None

    def validate(self, string, position):
        previous = self.previous
        self.previous = string

        if string.count(".") > 1:
            # if user types "." and there is already a dot, let the cursor
            # move to the other side
            if not string[position-1] == ".":
                # user pressed "." another time -> don't go forward
                position -= 1
            a, b = string.split(".", 1)
            string = ".".join([a, b.replace(".", "")])

            return self.Acceptable, string, position
        elif (previous is not None and previous.count(".")
                and not string.count(".") and position == previous.index(".")):
            # make sure removing decimal point does not lead to large numbers
            # (1.003 -> 1003)
            string = string[:position]
            return self.Acceptable, string, position
        elif string == "":
            return self.Intermediate, string, position
        elif string and string[position-1] in '.e-+':
            # remove trailing numbers
            return self.Intermediate, string.rstrip("0"), position
        elif valid_float_string(string):
            return self.Acceptable, string, position
        else:
            return self.Invalid, string, position

    def fixup(self, text):
        try:
            text = convert_float_to_string(convert_string_to_float(text))
        except ValueError:
            text = ""
        return text


def valid_float_string(string):
    try:
        convert_string_to_float(string)
    except ValueError:
        valid = False
    else:
        valid = True
    return valid


def convert_float_to_string(value):
    """Convert float to a string"""
    text = f"{value:.7g}"
    if not text.count("e"):
        # We don't have the exponential notation.
        text = str(value)
    return text


def convert_string_to_float(string):
    """Convert string of float or +/- "inf" to float"""
    try:
        val = float(string)
    except ValueError:
        string = string.strip()
        if string.startswith("-"):
            asign = -1
        else:
            asign = 1
        string = string.strip("+-")
        for iid in ["i", "in", "inf"]:
            if string == iid:
                val = asign * np.inf
                break
        else:
            raise ValueError("Not a valid float!")
    return val



# flake8: noqa: E402 (matplotlib.use has to be right after the import)
import pathlib
import pkg_resources
import signal
import sys
import traceback
import webbrowser

import matplotlib
matplotlib.use('QT5Agg')

from PyQt5 import uic, QtCore, QtWidgets

import afmformats
import h5py
import lmfit
import nanite
import nanite.read
import numpy
import scipy
import sklearn

from . import custom_widgets
from .dlg_tool_convert import ConvertDialog

from .. import registry
from .._version import version as __version__


class PyJibeQMdiSubWindow(QtWidgets.QMdiSubWindow):
    def closeEvent(self, QCloseEvent):
        """Correctly de-register a data set before removing the subwindow"""
        mainwidget = self.mdiArea().parentWidget().parentWidget()
        mainwidget.rem_subwindow(self.windowTitle())
        super(PyJibeQMdiSubWindow, self).closeEvent(QCloseEvent)


class PyJibe(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        """Initialize PyJibe

        If you pass the "--version" command line argument, the
        application will print the version after initialization
        and exit.
        """
        super(PyJibe, self).__init__(*args, **kwargs)
        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        QtCore.QCoreApplication.setOrganizationName("AFM-Analysis")
        QtCore.QCoreApplication.setOrganizationDomain("pyjibe.mpl.mpg.de")
        QtCore.QCoreApplication.setApplicationName("PyJibe")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
        #: PyJibe settings
        self.settings = QtCore.QSettings()
        self.settings.setIniCodec("utf-8")

        # load ui files
        path_ui = pkg_resources.resource_filename("pyjibe.head", "main.ui")
        uic.loadUi(path_ui, self)

        self.setWindowTitle("PyJibe {}".format(__version__))
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # Connect menu entries
        # File menu
        self.action_open_bulk.triggered.connect(self.on_open_bulk)
        self.action_open_single.triggered.connect(self.on_open_single)
        self.action_open_multiple.triggered.connect(self.on_open_multiple)
        # Edit menu
        is_dev_mode = bool(int(self.settings.value("developer mode", "0")))
        self.action_developer_mode.setChecked(is_dev_mode)
        self.on_developer_mode(is_dev_mode)
        self.action_developer_mode.triggered.connect(self.on_developer_mode)
        # Tool menu
        self.actionConvert_AFM_data.triggered.connect(self.on_tool_convert)
        # Help menu
        self.actionDocumentation.triggered.connect(self.on_documentation)
        self.actionSoftware.triggered.connect(self.on_software)
        self.actionAbout.triggered.connect(self.on_about)

        self.subwindows = []
        self.subwindow_data = []
        self.mdiArea.cascadeSubWindows()
        self.showMaximized()
        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(__version__)
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.AllEvents, 300)
            sys.exit(0)
        self.show()
        self.raise_()
        self.activateWindow()
        self.setWindowState(QtCore.Qt.WindowState.WindowActive)

    def load_data(self, files, retry_open=None, separate_analysis=False):
        """Load AFM data"""
        # expand directories
        data_files = []
        for ff in files:
            data_files += afmformats.find_data(ff)

        if not data_files:
            ret = QtWidgets.QMessageBox.warning(
                self,
                "No AFM data found!",
                "No AFM data files could be found in the location specified.",
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Retry
            )
            if retry_open is not None and ret == QtWidgets.QMessageBox.Retry:
                retry_open()
        else:
            # Make sure there are no duplicate files (#12)
            data_files = sorted(set(data_files))
            if separate_analysis:
                # open each file in one analysis
                usable = [[ss] for ss in data_files]
            else:
                usable = [data_files]
            for flist in usable:
                aclass = registry.fd.UiForceDistance
                self.add_subwindow(aclass, flist)

    def add_subwindow(self, aclass, flist):
        """Add a subwindow, register data set and add to menu"""
        sub = PyJibeQMdiSubWindow()
        inst = aclass(sub)
        sub.setWidget(inst)
        inst.add_files(flist)
        self.mdiArea.addSubWindow(sub)
        sub.show()
        self.subwindows.append(sub)
        # Add export choices
        if hasattr(inst, "get_export_choices"):
            choices = inst.get_export_choices()
            menobj = self.menuExport.addMenu(sub.windowTitle())
            for choice in choices:
                action = menobj.addAction(choice[0])
                action.triggered.connect(getattr(inst, choice[1]))

    def rem_subwindow(self, title):
        """De-register a data set and remove from the menu"""
        for ii, sub in enumerate(self.subwindows):
            if sub.windowTitle() == title:
                self.subwindows.pop(ii)
                break

        for action in self.menuExport.actions():
            if action.text() == title:
                self.menuExport.removeAction(action)
                break

    @QtCore.pyqtSlot()
    def on_about(self):
        about_text = "PyJibe is a user interface for data analysis in " \
            + "atomic force microscopy with an emphasis on biological " \
            + "specimens, such as tissue sections or single cells.\n\n" \
            + "Author: Paul Müller\n" \
            + "Repository: https://github.com/AFM-analysis/PyJibe\n" \
            + "Documentation: https://pyjibe.readthedocs.io"
        QtWidgets.QMessageBox.about(self,
                                    "PyJibe {}".format(__version__),
                                    about_text)

    @QtCore.pyqtSlot(bool)
    def on_developer_mode(self, checked):
        # remember in settings
        self.settings.setValue("developer mode", str(int(checked)))
        # set nanite
        if checked:
            nanite.read.DEFAULT_MODALITY = None
        else:
            nanite.read.DEFAULT_MODALITY = "force-distance"

    @QtCore.pyqtSlot()
    def on_documentation(self):
        webbrowser.open("https://pyjibe.readthedocs.io")

    @QtCore.pyqtSlot()
    def on_open_bulk(self):
        dlg = custom_widgets.DirectoryDialogMultiSelect(self)
        search_dir = self.settings.value("paths/load data", "")
        dlg.setDirectory(search_dir)
        if dlg.exec_():
            files = dlg.selectedFiles()
            if files:
                self.load_data(files=files, retry_open=self.on_open_bulk,
                               separate_analysis=False)
                self.settings.setValue("paths/load data",
                                       str(dlg.getDirectory()))

    @QtCore.pyqtSlot()
    def on_open_multiple(self):
        dlg = custom_widgets.DirectoryDialogMultiSelect(self)
        search_dir = self.settings.value("paths/load data", "")
        dlg.setDirectory(search_dir)

        if dlg.exec_():
            files = dlg.selectedFiles()
            if files:
                self.load_data(files=files, retry_open=self.on_open_multiple,
                               separate_analysis=True)
                self.settings.setValue("paths/load data",
                                       str(dlg.getDirectory()))

    @QtCore.pyqtSlot()
    def on_open_single(self):
        ext_opts = []
        # all
        exts = ["*"+e for e in registry.known_suffixes]
        ext_opts.append("Supported file types ({})".format(" ".join(exts)))
        # individual
        for suffix in registry.known_suffixes:
            for item in afmformats.formats.formats_by_suffix[suffix]:
                ext_opts.append("{} - {} (*{})".format(
                    item["maker"], item["descr"], suffix))
        exts_str = ";;".join(ext_opts)

        search_dir = self.settings.value("paths/load data", "")
        n, _e = QtWidgets.QFileDialog.getOpenFileNames(
            self, "Open single file", search_dir, exts_str, "")
        if n:
            # user did not press cancel
            self.load_data(files=n, retry_open=self.on_open_single)
            self.settings.setValue("paths/load data",
                                   str(pathlib.Path(n[0]).parent))

    @QtCore.pyqtSlot()
    def on_software(self):
        libs = [afmformats,
                h5py,
                lmfit,
                matplotlib,
                nanite,
                numpy,
                sklearn,
                scipy,
                ]
        sw_text = "PyJibe {}\n\n".format(__version__)
        sw_text += "Python {}\n\n".format(sys.version)
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += "- {} {}\n".format(lib.__name__, lib.__version__)
        sw_text += "- PyQt5 {}\n".format(QtCore.QT_VERSION_STR)
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self,
                                          "Software",
                                          sw_text)

    @QtCore.pyqtSlot()
    def on_tool_convert(self):
        dlg = ConvertDialog(self)
        dlg.show()


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    Parameters
    ----------
    etype: Exception
        exception type (`SyntaxError`, `ZeroDivisionError`, etc...)
    value: str
        exception error message
    trace:
        traceback header, if any (otherwise, it prints the
        standard Python header: ``Traceback (most recent call last)``
    """
    vinfo = "Unhandled exception in PyJibe version {}:\n".format(__version__)
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



"""Widget containing preprocessing plot"""
from PyQt5 import QtWidgets

from .mpl_preproc import MPLPreproc


class WidgetPlotPreproc(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        """Display preprocessing graph with navigation"""
        super(WidgetPlotPreproc, self).__init__(*args, **kwargs)

        self.mplvl = QtWidgets.QVBoxLayout(self)
        # Setup the matplotlib interface for approach retract plotting
        self.mpl_curve = MPLPreproc()
        self.mpl_curve.add_toolbar(self)
        self.mplvl.addWidget(self.mpl_curve.canvas)
        self.mplvl.addWidget(self.mpl_curve.toolbar)

    @property
    def fd(self):
        return self.parent().parent().parent()

    def update_details(self, details):
        """Update UI with details"""
        self.mpl_curve.clear()
        if details is None:
            return
        elif "correct_tip_offset" in details:
            meth = details["correct_tip_offset"]
            self.mpl_curve.clear()
            self.mpl_curve.axis.set_title(meth["method"])
            for key in meth:
                if key.startswith("plot "):
                    self.mpl_curve.axis.plot(*meth[key], label=key[5:])
            self.mpl_curve.axis.legend()
        self.mpl_curve.canvas.draw()



"""Test of data set functionalities"""
from PyQt5 import QtWidgets, QtCore

import pyjibe.head

from helpers import make_directory_with_data


def test_qmap_with_unused_curves(qtbot):
    """Uncheck "use" in the curve list and switch to the qmap tab"""
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data())
    war = main_window.subwindows[0].widget()
    # uncheck first curve
    cl1 = war.list_curves.currentItem()
    cl1.setCheckState(3, QtCore.Qt.Unchecked)
    war.tabs.setCurrentIndex(5)
    QtWidgets.QApplication.processEvents()



import pkg_resources

from nanite import preproc
from PyQt5 import QtCore, QtWidgets, uic


class WidgetPreprocessItem(QtWidgets.QWidget):
    #: Triggered whenever the state changes
    preproc_step_changed = QtCore.pyqtSignal()

    def __init__(self, identifier, *args, **kwargs):
        """Special widget for preprocessing options"""
        super(WidgetPreprocessItem, self).__init__(*args, **kwargs)

        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "widget_preprocess_item.ui")
        uic.loadUi(path_ui, self)

        # set label text
        name = preproc.get_name(identifier)
        self.label.setText(name)

        self.identifier = identifier

        # set tooltip
        meth = preproc.get_func(identifier)
        self.setToolTip(meth.__doc__)

        # set options
        if meth.options is not None:
            for opt in meth.options:
                choices = opt.get("choices", [])
                choices_hr = opt.get("choices_human_readable", choices)
                for text, data in zip(choices_hr, choices):
                    self.comboBox.addItem(text, data)
        else:
            self.comboBox.hide()

        # initially set-up enabled/disabled
        self.update_enabled()

        # signal: passthrough stateChanged
        self.checkBox.stateChanged.connect(self.on_state_changed)
        self.comboBox.currentIndexChanged.connect(self.on_state_changed)

    def get_options(self):
        """Return preprocessing options"""
        popts = {}
        if self.comboBox.isEnabled() and not self.comboBox.isHidden():
            meth = preproc.get_func(self.identifier)
            if meth.options is not None:
                for opt in meth.options:
                    if "choices" in opt:
                        popts[opt["name"]] = self.comboBox.currentData()
        return popts

    def isChecked(self):
        return self.checkBox.isChecked()

    @QtCore.pyqtSlot()
    def on_state_changed(self):
        self.update_enabled()
        self.preproc_step_changed.emit()

    def setChecked(self, *args, **kwargs):
        self.checkBox.setChecked(*args, **kwargs)

    @QtCore.pyqtSlot()
    def update_enabled(self):
        if self.isChecked():
            self.widget.setEnabled(True)
        else:
            self.widget.setEnabled(False)



# flake8: noqa: F401
from .dirdialog_multiselect import DirectoryDialogMultiSelect
from .mpl_navigation_toolbar_icons import (
    NavigationToolbarIndent,
    NavigationToolbarEDelta,
    NavigationToolbarQMap,
    NavigationToolbarPreproc
    )
from .wait_cursor import ShowWaitCursor, show_wait_cursor



import pathlib
import shutil
import tempfile

data_dir = here = pathlib.Path(__file__).parent / "data"


def make_directory_with_data(num_files=1):
    tdir = pathlib.Path(tempfile.mkdtemp(prefix="pyjibe_test_"))
    files = []
    for ii in range(num_files):
        target = tdir / f"test_{ii}.jpk-force"
        shutil.copy2(data_dir / "spot3-0192.jpk-force", target)
        files.append(target)
    return files



import pkg_resources

from nanite import preproc
from PyQt5 import uic, QtCore, QtWidgets

from .widget_preprocess_item import WidgetPreprocessItem


class TabPreprocess(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabPreprocess, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_preprocess.ui")
        uic.loadUi(path_ui, self)

        # Setup everything necessary for the preprocessing tab:
        # Get list of preprocessing methods
        premem = preproc.available()

        self._map_widgets_to_preproc_ids = {}
        for pid in premem:
            pwidget = WidgetPreprocessItem(identifier=pid, parent=self)
            self._map_widgets_to_preproc_ids[pwidget] = pid
            self.layout_preproc_area.addWidget(pwidget)
            if pid == "correct_tip_offset":
                idx = pwidget.comboBox.findData("deviation_from_baseline")
                pwidget.comboBox.setCurrentIndex(idx)
            pwidget.preproc_step_changed.connect(self.on_preproc_step_changed)
        spacer_item = QtWidgets.QSpacerItem(20, 0,
                                            QtWidgets.QSizePolicy.Minimum,
                                            QtWidgets.QSizePolicy.Expanding)
        self.layout_preproc_area.addItem(spacer_item)

        # Add recommended item (see `self.preproc_set_preset`)
        self.cb_preproc_presel.addItem("Recommended")
        self.cb_preproc_presel.activated.connect(self.on_preset_changed)
        self.cb_preproc_presel.currentIndexChanged.connect(
            self.on_preset_changed)
        # Apply recommended defaults
        self.cb_preproc_presel.setCurrentIndex(1)

    @property
    def fd(self):
        return self.parent().parent().parent().parent()

    def apply_preprocessing(self, fdist=None):
        """Apply the preprocessing steps if required"""
        if fdist is None:
            if hasattr(self, "fd"):
                fdist = self.fd.current_curve
            else:
                # initialization not finished
                return
        identifiers, options = self.current_preprocessing()
        # Perform preprocessing
        preproc_visible = self.fd.stackedWidget.currentWidget() == \
            self.fd.widget_plot_preproc
        details = fdist.apply_preprocessing(identifiers,
                                            options=options,
                                            ret_details=preproc_visible)
        if preproc_visible:
            self.fd.widget_plot_preproc.update_details(details)

    @QtCore.pyqtSlot()
    def on_preproc_step_changed(self):
        self.check_selection()
        self.apply_preprocessing()

    @QtCore.pyqtSlot()
    def check_selection(self):
        """If the user selects an item, make sure requirements are checked"""
        sender = self.sender()
        state = sender.isChecked()
        if sender in self._map_widgets_to_preproc_ids:
            pid = self._map_widgets_to_preproc_ids[sender]
            if state:
                # Enable all steps that this step here requires
                req_stps = preproc.get_steps_required(pid)
                if req_stps:
                    for pwid in self._map_widgets_to_preproc_ids:
                        if self._map_widgets_to_preproc_ids[pwid] in req_stps:
                            pwid.setChecked(True)
            else:
                # Disable all steps that depend on this one
                for dwid in self._map_widgets_to_preproc_ids:
                    did = self._map_widgets_to_preproc_ids[dwid]
                    req_stps = preproc.get_steps_required(did)
                    if req_stps and pid in req_stps:
                        dwid.setChecked(False)

    def current_preprocessing(self):
        # Note: Preprocessing is cached once in `fdist`.
        # Thus calling this method a second time without any
        # change in the GUI is free.
        identifiers = []
        options = {}
        for pwidget in self._map_widgets_to_preproc_ids:
            pid = self._map_widgets_to_preproc_ids[pwidget]
            if pwidget.isChecked():
                identifiers.append(pid)
                popts = pwidget.get_options()
                if popts:
                    options[pid] = popts
        # Make sure the order is correct
        identifiers = preproc.autosort(identifiers)
        return identifiers, options

    @QtCore.pyqtSlot()
    def on_preset_changed(self):
        """Update preselection"""
        text = self.cb_preproc_presel.currentText()
        if text == "None":
            used_methods = []
        elif text == "Recommended":
            used_methods = ["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset",
                            "correct_split_approach_retract"]
        else:
            raise ValueError(f"Unknown text '{text}'!")

        for pwidget in self._map_widgets_to_preproc_ids:
            pwidget.blockSignals(True)
            pid = self._map_widgets_to_preproc_ids[pwidget]
            pwidget.setChecked(pid in used_methods)
            pwidget.blockSignals(False)
        self.apply_preprocessing()

    def set_preprocessing(self, preprocessing, options=None):
        """Set preprocessing (mostly used for testing)"""
        if options is None:
            options = {}
        for pwidget in self._map_widgets_to_preproc_ids:
            pid = self._map_widgets_to_preproc_ids[pwidget]
            pwidget.setChecked(pid in preprocessing)
            if pid in options:
                opts = options[pid]
                key = sorted(opts.keys())[0]  # not future-proof
                idx = pwidget.comboBox.findData(opts[key])
                pwidget.comboBox.setCurrentIndex(idx)
